/*
 *  linktitlehandler.c
 *  Handlers for <link> and <title>
 *
 *  Created by Charles Mercadal on Tue Dec 09 2003.
 *
 *  This code is derived from software contributed to rss2html
 *  by Lyle Hanson.
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "common.h"
#include "cmfgetln.h"
#include "parsefunctions.h"

#define TMP_PATH "/tmp/.rss2html_bufferXXXXXX"
 
static void printLinkTitle(char *printFirst);
static void bufferOutput(void);
static void resumeOutput(void);
static void printSavedBuffer(void);
static void savePreTitle(char *data);

void linkHandler(char *);
void titleHandler(char *);


enum { NO = 0, IN_PROGRESS = 1, DONE = 2 };
static int saved_stdout, new_stdout;
static int buffering = NO;
static char *preTitleBuf = NULL;  /* Holds tags for hyperlinked title */
static char *linkloc     = NULL;
static char *title       = NULL;
static char tmp_filename[] = TMP_PATH;



void titleHandler(char *print)
{
    char *minbuf;


    minbuf = cmfgettoch(stdin, '<');
    title = xstrdup(minbuf);
    free(minbuf);
    
    if (linkloc != NULL)
        printLinkTitle(print);
    else {  /* Need to find the matching link for this title */
        savePreTitle(print);
        bufferOutput();
    }

    parseTag(); /* parse the closing tag that's left */

    if (buffering == DONE)
        printSavedBuffer();
}


void linkHandler(char *print)
{
    char *minbuf;


    minbuf = cmfgettoch(stdin, '<');
    linkloc = xstrdup(minbuf);
    free(minbuf);

    if (title != NULL)
        printLinkTitle(print);
    else {  /* Need to find the matching title for this link */
        savePreTitle(print);
        bufferOutput();
    }
     
    parseTag(); /* parse close tag left */

    if (buffering == DONE)
        printSavedBuffer();
}


void printLinkTitle(char *printFirst)
{

    if (buffering)
        resumeOutput(); /* Re-enable stdout */

    if (preTitleBuf) {
        printf("%s", preTitleBuf);
        free(preTitleBuf);
        preTitleBuf = NULL;
    }

    printf("%s", printFirst);

    (void)printf("<a href=\"%s\">%s</a>\n", linkloc, title);
    free(linkloc);
    free(title);
    
    /* Reset, linkHandler & titleHandler assume a href is printed */
    linkloc = title = NULL;
    buffering = DONE;
}


/*
 * We found one of the <title> or <link> tags, so we'll temporarily store other 
 * output until we find the matching tag.  This way, we can print the title
 * hyperlink before the rest of the output, and things are ordered like they
 * should be.  (LH)
 */
void bufferOutput(void)
{
    int file_d;
    FILE *tmpFile;

    if (buffering) {
        (void)fprintf(stderr, "ERROR: Title/Link matching doesn't support "
                      "nested tags (yet?).\n");
        return;
    }
    
    fflush(stdout);
    saved_stdout = dup(fileno(stdout));
    if (saved_stdout == -1) {
        perror("Saving stdout file descriptor");
        return;
    }
    
    file_d = mkstemp(tmp_filename);
    if (file_d == -1) {
        perror("Opening temporary file");
    }

    tmpFile = fdopen(file_d, "w");
    if (!tmpFile) {
        perror("Associating stream with temp file descriptor");
        return;
    }
    
    new_stdout = fileno(tmpFile);
    if (new_stdout == -1) {
        perror("Getting descriptor for temp file");
        return;
    }
    
    dup2(new_stdout, fileno(stdout));
    if (new_stdout == -1) {
        perror("Redirecting stdout to temp file");
        return;
    }

    buffering = IN_PROGRESS;
}


/*
 * After we've found matching <title> and <link> tags and have printed the
 * hyperlinked title, restore stdout and print the data we've been buffering.
 * (LH)
 */
void resumeOutput()
{

    fflush(stdout);
    dup2(saved_stdout, fileno(stdout)); /* Restore the old stdout */
    buffering = NO;
}


/* Prints contents of the temporary file to the newly-restored stdout. (LH) */
void printSavedBuffer(void)
{
    register int ch;
    FILE *saved_data;
    

    saved_data = fopen(tmp_filename, "r");
    /* Read saved data and print to stdout */
    while ((ch = fgetc(saved_data)) != EOF)
        (void)putchar(ch);
    
    if (remove(tmp_filename))
        perror("Removing temp file");

    /* Restore tempfile name template */
    strcpy(tmp_filename, TMP_PATH);
    buffering = NO;
}


/*
 * When we find a <title> or <link> tag, and we have some html to print before
 * the complete hyperlinked title, we need to save this data so that we can
 * print it BEFORE the linked title and then the buffered data in tmp_buffer.
 * (LH)
 */
void savePreTitle(char *data)
{
    char *tmp;
    int first_time = 0;


    if (preTitleBuf == NULL)
        first_time = 1;
    
    tmp = xrealloc(preTitleBuf, strlen(data) + 1);
    
    preTitleBuf = tmp;
    
    if (first_time)
        preTitleBuf[0]='\0';
    
    strcat(preTitleBuf, data);
}
